#include <cuda.h>
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <cstdio>
#include <cstdlib>
#include <malloc.h>

static size_t DEVICE_ELEMENTS = 256*1024*1024;
static size_t LAYER_ELEMENTS = 512*512;
static size_t LAYERS = 16*1024;

static size_t MEMORY_ALIGNMENT = 4096;

#define CUDA_SAFE_CALL( call ) \
{ \
    const cudaError_t error = call; \
    if( error != cudaSuccess ) \
    { \
        printf( "Error: %s: %d, ", __FILE__, __LINE__ ); \
        printf( "code: %d, reason: %s\n", error, cudaGetErrorString( error ) ); \
        exit( 1 ); \
    } \
} 

__global__ void doStuff( float* data, size_t n )
{
    auto idx = blockIdx.x*blockDim.x+threadIdx.x;
    if ( idx < n )
        data[idx] += idx;
}

int main(int argc, char **argv)
{
    int iterations = 1;
    int gpuId = 0;

    if ( argc >= 2 )
        iterations = atoi(argv[1]);

    if ( argc >= 3 )
        gpuId = atoi(argv[2]);

    printf("CUDA 980Ti driver crash test.\nThe test should result in a BSOD with drivers newer than 378.92\n");

    int devices;
    CUDA_SAFE_CALL( cudaGetDeviceCount(&devices) );

    printf("Using device #%d (of %d)\n", gpuId, devices);
    CUDA_SAFE_CALL( cudaSetDevice( gpuId ) );

    cudaDeviceProp prop;
    CUDA_SAFE_CALL( cudaGetDeviceProperties( &prop, gpuId ) );

    printf("Device name: %s, total mem: %f GiB\n", prop.name, prop.totalGlobalMem/1024.0f/1024.0f/1024.0f);

    float** hostData = new float*[LAYERS];

    printf("Allocating %f GiB on the host (%lu layers)...\n", LAYERS*LAYER_ELEMENTS*sizeof(float)/1024.0f/1024.0f/1024.0f, LAYERS );
    for (size_t i = 0; i < LAYERS; i++)
        hostData[i] = reinterpret_cast<float*>( malloc( LAYER_ELEMENTS*sizeof(float) ) );
        //hostData[i] = reinterpret_cast<float*>( _aligned_malloc( LAYER_ELEMENTS*sizeof(float), MEMORY_ALIGNMENT ) );

    printf("Initializing data...\n" );
    for (size_t i = 0; i < LAYERS; i++)
        for (size_t j = 0; j < LAYER_ELEMENTS; j++)
            hostData[i][j] = 42.0f;

    for ( int k = 0; k < iterations; k++ )
    {
        printf("Running mapped memory access tests %d/%d...\n", (k+1), iterations );

        for (size_t i = 0; i < LAYERS; i++)
        {
            float* devicePtr;

            //CUDA_SAFE_CALL( cudaHostRegister( hostData[i], LAYER_ELEMENTS*sizeof(float), CU_MEMHOSTALLOC_DEVICEMAP | CU_MEMHOSTALLOC_PORTABLE ) );
            CUDA_SAFE_CALL( cudaHostRegister( hostData[i], LAYER_ELEMENTS*sizeof(float), cudaHostRegisterMapped | cudaHostRegisterPortable ) );

            CUDA_SAFE_CALL( cudaHostGetDevicePointer( (void **)&devicePtr, (void *)hostData[i], 0 ) );

            doStuff<<< ceil(LAYER_ELEMENTS/1024.0) , 1024 >>>( devicePtr, LAYER_ELEMENTS );

            CUDA_SAFE_CALL( cudaDeviceSynchronize() );

            CUDA_SAFE_CALL( cudaHostUnregister( hostData[i] ) );
        }

    }

    printf("Finished tests.\n" );

    for (size_t i = 0; i < LAYERS; i++)
        //_aligned_free( hostData[i] );
        free( hostData[i] );

    delete[]( hostData );

    return 0;
}
